#include <signal.h>
#include <unistd.h>
#include <string.h>
#include <errno.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <sys/ioctl.h>
#include <sys/socket.h>

#include "ipxd.h"
#include "ipxripd.h"
#include "ipxsapd.h"

int n_of_ifaces;
struct ipx_interface ifaces[MAX_IFACE];
int time_since_last_bcast=0;
static struct rip_packet rip_in_buf;
static struct sap_packet sap_in_buf;

volatile int dump_request=0;
volatile int timer_request=0;
volatile int terminate=0;

int new_log_entry=1;
int debug_option=0;
FILE *log_file=stderr;

static void
int_handler()
{
	signal(SIGINT,int_handler);
	dump_request=1;
}

static void
timer_handler()
{
	signal(SIGALRM,timer_handler);
	timer_request=1;
}

static void
terminate_handler()
{
	signal(SIGTERM,terminate_handler);
	terminate=1;
}

static void
hup_handler()
{
	signal(SIGHUP,hup_handler);
	terminate=2;
}

static int
ipx_create_ifc(IPXNet network, IPXNode node,
		      char *device, int type, void *data)
{
	struct ipx_interface *ifc = &(ifaces[n_of_ifaces]);

	if (n_of_ifaces >= MAX_IFACE)
	{
		LOG_START;
		fprintf(log_file,"too many interfaces (max. %i)\n",
			(int)MAX_IFACE);
		LOG_END;
		exit(1);
	}

	if (network == 0)
	{
		LOG_START;
		fprintf(log_file,"interface bound to net 0\n");
		LOG_END;
		exit(1);
	}
	ipx_assign_node(ifc->ifnode, node);

	if (ipx_rip_init_ifc(ifc, network, device, type, data) != 0)
	{
		exit(1);
	}
	if (ipx_sap_init_ifc(ifc, network, device, type, data) != 0)
	{
		exit(1);
	}
	n_of_ifaces += 1;
	return 0;
}

void
init()
{
	struct itimerval itval;
	
	LOG_ENTRY;
	LOG_START;
	fprintf(log_file,"Init start\n");
	LOG_END;
	
	n_of_ifaces = 0;

	ipx_kern_scan_ifaces(ipx_create_ifc, NULL);

	itval.it_interval.tv_sec = TIMER_RATE;
	itval.it_value.tv_sec = TIMER_RATE;
	itval.it_interval.tv_usec = 0;
	itval.it_value.tv_usec = 0;
	if (setitimer(ITIMER_REAL, &itval, (struct itimerval *)NULL) < 0)
	{
		LOG_START;
		fprintf(log_file, "FATAL ERROR: can't set itimer: %s\n",
			strerror(errno));
		LOG_END;
		exit(1);
	}
	
	signal(SIGINT,int_handler);
	signal(SIGALRM,timer_handler);
	signal(SIGTERM,terminate_handler);
	signal(SIGHUP,hup_handler);

	ipx_rip_initial_broadcasts();
	ipx_sap_initial_broadcasts();
	
	LOG_ENTRY;
	LOG_START;
	fprintf(log_file,"Init end\n");
	LOG_END;
}

static void
run(void)
{
	struct sockaddr_ipx sipx;
	int addr_len=sizeof(sipx);
	int size;
	fd_set rset;
	int i, result;
	int max_sk = 0;
	struct ipx_interface *ifc = NULL;
	int got_rip, got_sap;

	while(1)
	{
		FD_ZERO(&rset);

		for (i = 0; i < n_of_ifaces; i += 1)
		{
			if (ifaces[i].r_output.sk > max_sk)
			{
				max_sk = ifaces[i].r_output.sk;
			}
			FD_SET(ifaces[i].r_output.sk, &rset);

			if (ifaces[i].s_output.sk > max_sk)
			{
				max_sk = ifaces[i].s_output.sk;
			}
			FD_SET(ifaces[i].s_output.sk, &rset);
		}		

		LOG_ENTRY;

		result = select(max_sk+1, &rset, NULL, NULL, NULL);

		if (   (result < 0)
		    && (errno != EINTR))
		{
			LOG_START;
			fprintf(log_file, "select error: %s\n",
				strerror(errno));
			LOG_END;
			continue;
		}

		size = -1;

		if (result == 0)
		{
			LOG_START;
			fprintf(log_file, "select timeout expired\n");
			LOG_END;
			continue;
		}

		got_rip = got_sap = 0;

		if (result > 0)
		{
			for (i = 0; i < n_of_ifaces; i += 1)
			{
				if (FD_ISSET(ifaces[i].r_output.sk, &rset))
				{
					ifc = &(ifaces[i]);
					size=recvfrom(ifaces[i].r_output.sk,
						      &rip_in_buf,
						      sizeof(rip_in_buf),
						      0,
						      (struct sockaddr*)&sipx,
						      &addr_len);
					got_rip = 1;
					/* only one packet at a time */
					break;
				}
				if (FD_ISSET(ifaces[i].s_output.sk, &rset))
				{
					ifc = &(ifaces[i]);
					size=recvfrom(ifaces[i].s_output.sk,
						      &sap_in_buf,
						      sizeof(sap_in_buf),
						      0,
						      (struct sockaddr*)&sipx,
						      &addr_len);
					got_sap = 1;
					/* only one packet at a time */
					break;
				}
			}
		}

		if (dump_request)
		{
			fdump_routes(log_file);
			fdump_servers(log_file);
			dump_request=0;
		}
		if (timer_request)
		{
			static int rip_broadcast = 0;
			int do_broadcast = 0;

			time_since_last_bcast += TIMER_RATE;
			if (time_since_last_bcast >= BROADCAST_TIME/2)
			{
				do_broadcast = 1;
				rip_broadcast = (rip_broadcast==0) ? 1 : 0;
				time_since_last_bcast -= BROADCAST_TIME/2;
			}

			LOG_ENTRY;
			ipx_rip_do_aging(TIMER_RATE,
					    (do_broadcast  != 0)
					 && (rip_broadcast != 0));
			LOG_ENTRY;
			ipx_sap_do_aging(TIMER_RATE,
					    (do_broadcast!=0)
					 && (rip_broadcast == 0));
			timer_request=0;
		}
		if (terminate)
		{
			LOG_START;
			if (terminate==1)
				fprintf(log_file,
					"exiting on signal 15 (SIGTERM)\n");
			else
				fprintf(log_file,
					"exiting on signal 1 (SIGHUP)\n");
			LOG_END;
			return;
		}
		if (size==-1)
		{
			continue;
		}
		if (got_rip != 0)
		{
			DL_START;
			fprintf(log_file,"RIP from: ");
			ipx_fprint_saddr(log_file,&sipx);
			fprintf(log_file,"\n");
			ipx_rip_fdump(log_file, &rip_in_buf, size);
			DL_END;
			handle_rip(&rip_in_buf, size, &sipx,ifc);
		}
		if (got_sap != 0)
		{
			DL_START;
			fprintf(log_file,"SAP from: ");
			ipx_fprint_saddr(log_file,&sipx);
			fprintf(log_file,"\n");
			ipx_sap_fdump(log_file, &sap_in_buf, size);
			DL_END;
			handle_sap(&sap_in_buf, size, &sipx,ifc);
		}	
	}
}

static void
help()
{
	fprintf(stderr,
		"IPX routing daemon v0.91, (c) Ales Dryak, 1995\n"
		"Usage: ipxripd [-d] [<path to log file>]\n");
}

static int
parse_cmdline(int argc,char** argv)
{
	char* log_name;
	argc--,argv++;
	for(; *argv && **argv=='-'; argc--,argv++)
	{
		if (*(*argv+1)=='d')
			debug_option=1;
		else
		{
			fprintf(stderr,
				"Unknown command line option %s\n",*argv);
			help();
			exit(1);
		}
	}
	if (argc>1)
	{
		fprintf(stderr,"Too many parameters");
		help();
		exit(1);
	}
	if (argc==0 && debug_option)
	{
		log_file=stdout;
		return 0;
	}
	log_name=DEFAULT_LOGNAME;
	if (argc==1) log_name=*argv;
	if ((log_file=fopen(log_name,"a"))==NULL)
	{
		perror(log_name);
		exit(1);
	}
	return 1;
}

static void
daemonize()
{
	int fd,c;
	
	if ((c = fork()) > 0) exit(0);
  	if (c < 0)
  	{
		fprintf(stderr, "ipxripd: can't fork: %s\n",strerror(errno));
		exit(1);
	}

	close(0);
	close(1);
	close(2);
	if ((fd = open("/dev/tty", O_RDWR)) >= 0) 
	{
		ioctl(fd, TIOCNOTTY, NULL);
		close(fd);
	}
}

int
main(int argc,char** argv)
{
	if (parse_cmdline(argc,argv)) daemonize();
	init();
	run();
	ipx_rip_done();
	ipx_sap_done();
	return 0;
}

